	#include <version_id.h>
/* ID_OF(XXX $Id: smbfiling.c,v 1.11 2009/06/24 13:42:14 thammer Exp $) */
/*============================================================================
** Copyright (c)  2001 - 2013 Xerox Corporation. All Rights Reserved.
** Copyright protection claimed includes all forms and matters of
** copyrightable material and information now allowed by statutory or
** judicial law or hereinafter granted, including without limitation,
** material generated from the software programs which are displayed
** on the screen such as icons, screen and the like.
*/
/*============================================================================
** FILE: smbfiling.c
**
** DESCRIPTION: This file defines functions that make system calls to the
**              SAMBA smbclient to perform filing operations on a remote
**              SMB server. Implemented for SMB scan-to-file.
**
**
**
** REVISION HISTORY:
**
** AUTHOR       DATE     DESCRIPTION
** ilakshmi     11/20/13 CQGbl00485645  Unable to scan 2 folders deep when system account 
**					does not have rights to 1 folder deep
** R. Wilkie    02/12/13 CQGbl00429732: Fixed issue where LDAP domain force was over
**                       Writing valid domain settings.  Updated some logging as well
** K.Sitaram    13/02/12 CQGbl00357794 When scanning SMB with an authenticated user setup, 
**                       prepending the domain name to the credentials, the scan will fail 
**                       with "login Failures".
** P. Tolentino 05/30/11 CQGbl00317920 - domain name not sent when network scanning is used
** M. Hinojosa 02/24/11  CQGbl00298118 - Fix for Buffer overrun error when parsing the string
**                       used for changing directory.
** Y. Gao		02/11/11 CQGbl00290638 removed the 127 path length restriction in cdCommand
**                       as SMB supports more than 127 char
** Y. Gao		01/24/11 CQGbl00290638-291332 Fixed crashing and incorrect
**                       seqmentation parsing when path > 127
** M. Hinojosa 09/09/10	 Use kerberos cache path to set the Kerberos environment variable.
** M. Lucyshyn 09/09/10  CQGbl00212276 -- Pass Kerberos Cache Path to SMB DLM.
** M. Hinojosa	09/02/10  Extend logic of using kerbetized samba client with
**                       the KerberosFilingMode parameter.
** M. Lucyshyn  08/27/10 Added KerberosFilingMode parameter to 
** M. Hinojosa	08/04/10 Add kerbetized samba client call.
** Jacob Goh  03/24/10   CQGbl00188614: replaced \" to ' in  commandString in callSMBclient().
** YGao	      06/10/09   CQGb100140265 - Spar72775 Clone
**			 CDDub31697: Scan2Home cannot log in with accented pw
**                       CDDub32568: Scan2Home cannot log in with accented username.
** THammer    06/11/08   CQGbl76614: replace hardcoded paths with xpaths
**                                   lib calls
**                           removed SMB_CLIENT_CALL definition, use
**                           XPATH_NC_DLMS_DIR with rest of string in its place
** M. Lucyshyn  05/03/05 CDDua87041: Reduce debug level to 0 when
**                       smbclient_x output is redirected to a file for a
**                       directory listing. Set Samba Debug level to 5 when
**                       the debug macro is set to (1).
** M. Lucyshyn  02/02/04 CDDua68494: Upgrade smbclient 2.0.6 to 3.0.1
** M. Lucyshyn  09/11/03 CDDua65081: Add close() for fd passed to
**                       to smb_filing_login().
** R. Wilkie    10/23/02 CDDua53891: Added in check for NULL passwords.
** M. Lucyshyn  06/05/02 CDDua47398: Convert '\' to '/' or vice-versa as
**                       needed. Also improve handling of spaces in file
**                       names.
** M. Lucyshyn  02/04/02 CDDua42568: Added "bad character"support for SMB
**                       filing.
** M. Lucyshyn  12/20/01 CDDua40592 -- Corrected bug discovered in
**                       checkForSpaces() while Porting SMB DLM to Linux.
**                       The procedure was ignoring existing quotes.
**                       Also corrected return value from linux system()
**                       call.
** M. Lucyshyn  11/5/01  CDDua39034 -- Added SESS logging statements
**                       and modified error handling so that strings con-
**                       taining only blanks would not be accepted as file
**                       names. Revised error codes to avoid conflicts with
**                       the Samba smbclient.
** M. Lucyshyn  10/16/01 Creation
*****************************************************************************/



/*****************************************************************************
**
**  IMPLEMENTATION NOTES:
**
**
*****************************************************************************/


/****************************************************************************/
/*  Include Files                                                           */
/****************************************************************************/
#include <stdio.h>
#include <string.h>
#include <appipc.h>
#include <smbapi.h>
#include <appipc.h>
#include <sessLog.h>
#include <x_paths.h>
#include "smb_logging.h"
#include "securityapi.h"
#include "smbipc.h"
#include "cfgAttr.h"

#if defined (__linux__)
#include <sys/wait.h>
#endif

/****************************************************************************/
/*  Compile-Time Checks                                                     */
/****************************************************************************/


/****************************************************************************/
/*  Local Macro Definitions                                                 */
/****************************************************************************/
#define DEBUG (0)  /* change to (1) to redirect output from smbclient into
                    * a file  and to show log entries containing a password
                    * for debugging purposes. Note: It is also necessary to
                    * set the XEROX_SAMBA_DEBUG environmental variable on the
                    * device in order to turn on smbclient_x printf
                    * statements.
                    */
#if DEBUG
#define COMMANDDEBUG (1)  /* IF DEBUG is defined, we are going to want this one too */
#else
#define COMMANDDEBUG (0) /* Change to (1) to see the command lines used in the
                          * debug logs for smbclient   Good to use if you do not
                          * need ALL the information DEBUG has */
#endif


#define USERID "USERID"
#define configServerBaseDnArg "ldap.appendDnValue"
#define configServerLdapAppendDnArg "ldap.appendDnEnabled"
#define configServiceLdapAuthEnabled "authentication.LDAP.enabled"


/****************************************************************************/
/*  Local Constants                                                         */
/****************************************************************************/
enum smb_filing_states{
   SMB_FILING_IDLE,
   SMB_FILING_ACTIVE
};

enum smb_filing_states filingState;


/****************************************************************************/
/*  Local Types                                                             */
/****************************************************************************/
typedef struct{
   /* declare the service sting. Allow space for three / characters and \0 */
   char  service[ SMB_DLM_SERVER_LENGTH + SMB_DLM_VOLUME_LENGTH + 4];
   /* declare the U string, allow room for " -u" ,% character and \0 */
   char  Ustring[ ((SMB_DLM_USERNAME_LENGTH + SMB_DLM_PASSWORD_LENGTH + 4)*3)
                  + 1 ];
   char username[SMB_DLM_USERNAME_LENGTH + 1];
   char localPath[SMB_DLM_PATH_LENGTH + 3]; /* allow for quotes and \0 */
   char remotePath[SMB_DLM_PATH_LENGTH + 3];
   char outFilePath[SMB_DLM_PATH_LENGTH + 3];
   int port;
   KerberosFilingMode krbFlngMode;
   char kerberosCachePath[SMB_DLM_PATH_LENGTH + 1];
   
} FilingSessionData;

/****************************************************************************/
/*  External Globals                                                        */
/****************************************************************************/
extern ipc_queue_id_t   smb_queue_id;
extern Ipc_queue *      smb_queue;

/****************************************************************************/
/*  Filescope Globals                                                       */
/****************************************************************************/
FilingSessionData session;
char command[MAX_COMMAND_LENGTH + 1];
ALIGN_TYPE ipcData;

/****************************************************************************/
/*  Extern Function Declarations                                            */
/****************************************************************************/
extern bool_t convert_string_to_decimal(char *toConvertString,
                                        char *convertedString);

/****************************************************************************/
/*  Function Prototypes                                                     */
/****************************************************************************/
       void smb_filing_login(Ipc_message * msg_p);
       void smb_filing_logout(void);
       void smb_filing_putFile( Ipc_message * msg_p);
       void smb_filing_getFile( Ipc_message * msg_p);
       void smb_filing_deleteFile( Ipc_message * msg_p);
       void smb_filing_makeDir( Ipc_message * msg_p);
       void smb_filing_removeDir( Ipc_message * msg_p);
       void smb_filing_listDir( Ipc_message * msg_p);

static void clearFilingSessionData(void);
static bool_t isEmpty( char * aString);
static bool_t cdCommand( char * cd_string, char * path);
static int callSMBclient( char * commandScript);
static void checkForSpaces( char * path);
static bool_t noPath( char * directoryName, char * remotePath);
static void pathCheck( char * path);
static void slash_to_backslash( char * path);
static void backslash_to_slash( char * path);
static bool_t doWeUseTheLDAPDomain();


/****************************************************************************/
/*  External Function Definitions                                           */
/****************************************************************************/


/****************************************************************************/
/*  Other Local Macros                                                      */
/****************************************************************************/


/****************************************************************************/
/*  Local Function Definitions                                              */
/****************************************************************************/



/*****************************************************************************

* FUNCTION: smb_filing_login
*
* DESCRIPTION: This function sets up an SMB filing session within smbipc.
* It calls an utility routine to  initialize the filingSessionData structure.
* Using the file descriptor from the IPC message, it opens the file and reads
* the data into the filingSessionData structure.  It then attempts to connect
* to the SMB file server using a system call to smbclient.
* If the system call returns (0), this function sets the process state to
* SMB_FILING_ACTIVE, otherwise it sets the process state to SMB_FILING_IDLE.
* It then sends an IPC message back to the sender with message type
* SMB_LOGIN_REPLY and a zero error code for success or a non-zero error code
* for failure.
*
*
* PARAMETERS:
*              Ipc_message * msg_p -- pointer to IPC message
*
* RETURNS:     void
*****************************************************************************/
void smb_filing_login(Ipc_message * msg_p)
{
   int fd;
   int errorCode;
   ipc_queue_id_t return_qID;
   char * buf_ptr;
   char readBuf[MAX_SMB_FILING_DATA_SIZE + 1];
   char decUserName[SMB_DLM_USERNAME_LENGTH * 3 + 1];
   char decPassword[SMB_DLM_PASSWORD_LENGTH * 3 + 1];
   int bytes_read = 0;
   int total_bytes_read = 0;
   char * tmp;
   char cmdScript[25];
   size_t strLength;
   char *pBackslash = NULL, *pSlash = NULL, *pUsername = NULL;
   status_t status;
   bool_t stripDomain = FALSE;
   status_t cfgStatus;
   char attrValue[MAX_CONFIG_ITEM_SIZE];
   char tempLoginName[1024*2+1];
   char *domainNameFilter = NULL;

   LOG_SMB_ENTRY;

   clearFilingSessionData();
   return_qID = ipc_get_message_source( msg_p );

   /* Get the fd and get copy into filing Session data */
   fd = get_message_fd( msg_p);
   if ( fd < 0 )
   {
       errorCode = SMB_DLM_GENERIC_ERROR;
       ipcData = (ALIGN_TYPE)errorCode;
       ipc_send_message( return_qID, get_send_handle(), &ipcData,
                         sizeof(ipcData),SMB_DLM_LOGIN_REPLY);
       LOG_SMB_ERROR("A valid file descriptor was not sent");
       LOG_SMB_RETURN;
   }
   buf_ptr = readBuf;

   /* read in Data, borrowed from smbipc */
   while ((total_bytes_read < MAX_SMB_FILING_DATA_SIZE) &&
       ((bytes_read = read (fd, buf_ptr, MAX_SMB_FILING_DATA_SIZE -
                   total_bytes_read)) > 0))
   {
       total_bytes_read += bytes_read;
       buf_ptr += bytes_read;
   }

   close(fd); /* Close file descriptor created by get_message_fd() */

   if(bytes_read == -1)
   {
      LOG_SMB_ERROR("%s(): bytes read were -1", __FUNC__);
      errorCode = SMB_DLM_GENERIC_ERROR;
      ipcData = (ALIGN_TYPE)errorCode;
      ipc_send_message( return_qID, get_send_handle(), &ipcData,
                        sizeof(ipcData), SMB_DLM_LOGIN_REPLY);
      LOG_SMB_RETURN;
   }
   /* Copy in the data, get the server name and possibly the port number */
   /* Items in the buffer are delineated by '\0'*/

   readBuf[MAX_SMB_FILING_DATA_SIZE] = '\0'; /* just to be safe */

   buf_ptr = readBuf; /* reset the buffer pointer */

   /* Get the port number */

   strLength = strlen(buf_ptr);

   /* copy the port number into port */
   session.port = atoi(buf_ptr);

   /*
    * Get the user name or login name
    */

   buf_ptr += strLength + 1; /* advance pointer */
   strLength = strlen( buf_ptr);
   strcpy(session.Ustring, " -u");

   status = Cfg_GetBoolAttrValue("scanToHome.stripDomain", &stripDomain);
   if (status != X_SUCCESS)
   {
    LOGERROR("%s: Cfg_GetBoolAttrValue(%s) failed",
             __FUNCTION__, "scanToHome.stripDomain");
   }

   /*
    * WARNING WARNING, THIS IS A HACK, WE NEED TO MOVE THIS FUNCTIONALITY INTO THE SCAN SERVICES
    */
   if (stripDomain)
   {
       LOGFORCE("  Using Strip Domain Hack in SMB");

	/*
	  If the username has a backslash (or slash), strip everything before the backslash 
	  (not including the backslash itself) off the passed-in username.
	*/
	pBackslash = strrchr(buf_ptr, '\\');
	if (pBackslash != NULL)
	{
		pSlash = strrchr(pBackslash, '/');
		if (pSlash != NULL)
		{
			pUsername = pSlash;
		}
		else
		{
			pUsername = pBackslash;
		}
	}
	else
	{
		pSlash = strrchr(buf_ptr, '/');
		if (pSlash != NULL)
		{
			pUsername = pSlash;
		}
		else
		{
			pUsername = buf_ptr;
		}
	}
   }
   else
   {
   	pUsername = buf_ptr;
   }

   /*
    * WARNING WARNING, THIS IS A HACK, WE NEED TO MOVE THIS FUNCTIONALITY INTO THE SCAN SERVICES
    * If we are setup to use the USER PROVIDED ldap domain, do so, else skip this bit
    ******/
   if(doWeUseTheLDAPDomain() == TRUE)
   {
       LOGFORCE("  Using LDAP Authentication Domain Hack in SMB");

       cfgStatus = Cfg_GetAttrValue(configServerBaseDnArg, attrValue,MAX_CONFIG_ITEM_SIZE);
       if (cfgStatus != X_SUCCESS)
       {
           LOGERROR("Value for %s could not be retrieved from the Config Server",
                                        configServerBaseDnArg);
       }
       else
       {
           if(NULL != strstr(attrValue, USERID))
           {
               domainNameFilter = strstr(attrValue, USERID) + strlen(USERID)+1;
               if(domainNameFilter != NULL)
               {
                   sprintf(tempLoginName, "%s%s%s", domainNameFilter,"\\",pUsername);
                   pUsername = tempLoginName;
               }
               else
               {
                   LOGERROR("%s(): Domain name is empty",__FUNC__);
               }
           }
       }
   } /* if(doWeUseLDAPDomain() == TRUE) */


   strncpy(session.username,pUsername,SMB_DLM_USERNAME_LENGTH); //store username for future use
   session.username[SMB_DLM_USERNAME_LENGTH] = '\0';

   /* Convert the user name to a decimal string */
   if(!convert_string_to_decimal(pUsername, decUserName))
   {
      LOG_SMB_ERROR("%s(): Could not convert username \"%s\" to decimal", __FUNC__,
                   pUsername);
      errorCode = SMB_DLM_INTERNAL_ERROR;
      ipcData = (ALIGN_TYPE)errorCode;
      ipc_send_message( return_qID, get_send_handle(), &ipcData,
                        sizeof(ipcData), SMB_DLM_LOGIN_REPLY);
      LOG_SMB_RETURN;
   }

   LOGINFO(" Login for scanning username [%s] dec[%s]", pUsername, decUserName);

   strcat(session.Ustring, decUserName);
   strcat(session.Ustring, "%");

   /* Retrieve the value of the KerberosFilingMode enumeration */
   buf_ptr +=strLength + 1; /* advance pointer */
   session.krbFlngMode = atoi( buf_ptr );
   LOG_SMB_DEBUG("%s(): value of KerberosFilingMode enumerator = %d",
                  __FUNC__, session.krbFlngMode);

   /* Retreive the path to the kerberos cache for this user */
   strLength = strlen(buf_ptr);
   buf_ptr +=strLength + 1; /* advance the pointer */
   strcpy( session.kerberosCachePath, buf_ptr);
   LOG_SMB_DEBUG("%s(): Kerberos Cache Path = %s", __FUNC__,
                 session.kerberosCachePath);

   /* Get the password */
   tmp = (char *)ipc_message_data(msg_p);

#if DEBUG
   LOG_SMB_DEBUG("%s(): password as received from IPC message: %s", __FUNC__, tmp);
#endif

   /*
    * Check for a NULL (empty) password, if its not NULL, proceed to
    *  adding in the password to the command line for password/username
    *  combo.  Else if the password is null, we do not need to bother
    *  adding it to the command line.
    */
   if ((tmp != NULL) && (strlen(tmp) > 0))
   {
       /* convert the password into a decimal string */
       if(!convert_string_to_decimal(tmp, decPassword))
       {
          LOG_SMB_ERROR("%s(): Could not convert password to decimal",
                                                                  __FUNC__);
          errorCode = SMB_DLM_INTERNAL_ERROR;
          ipcData = (ALIGN_TYPE)errorCode;
          ipc_send_message( return_qID, get_send_handle(), &ipcData,
                            sizeof(ipcData), SMB_DLM_LOGIN_REPLY);
          LOG_SMB_RETURN;
       }
       strcat(session.Ustring, decPassword);
   }

   /*
   ** Construct the command script for the smbclient call. Make it a blank
   ** character so that no output is produced by smbclient_x, but it will
   ** still force smbclient_x to return.
   */
   strcpy(cmdScript, " ");

   /*
    * For Debugging, provide  redirect file
    */
#if DEBUG
   strcpy(session.outFilePath, "/var/log/nc/logInDebug");
#else
   session.outFilePath[0] = '\0';
#endif

   if(( errorCode = callSMBclient(cmdScript)) == 0)
   {
	   filingState = SMB_FILING_ACTIVE;
   }
   else
   {
	   filingState = SMB_FILING_IDLE;
	   clearFilingSessionData();
   }

   /* reply to the caller */
   ipcData = (ALIGN_TYPE)errorCode;
   ipc_send_message( return_qID, get_send_handle(), &ipcData, sizeof(ipcData),
                     SMB_DLM_LOGIN_REPLY);

#if DEBUG
   session.outFilePath[0] = '\0';
#endif
   LOG_SMB_RETURN;
 /*  return; */
}


/*****************************************************************************

* FUNCTION: smb_filing_logout
*
* DESCRIPTION: This function closes the filing session. It sets the
*              filingState to SMB_FILING_IDLE and clears the filing session
*              data structure.  The process sending the IPC message does not
*              expect a response.
*
*
* PARAMETERS:  none
*
* RETURNS:     void
*****************************************************************************/
void smb_filing_logout(void)
{
   LOG_SMB_ENTRY;
   filingState = SMB_FILING_IDLE;
   clearFilingSessionData();
   LOG_SMB_RETURN;
}


/*****************************************************************************

* FUNCTION: smb_filing_putFile
*
* DESCRIPTION: This function puts a file onto the file server. It expects
*              in the data section of the IPC message to find first the local
*              path and remote path seperated by '\0' in that exact order. The
*              path string must include the filename. It sends back to the
*              caller an IPC message informing caller about success or failure.
*
*
*
* PARAMETERS:  Ipc_message * msg_p
*
* RETURNS:     void
*****************************************************************************/
void  smb_filing_putFile( Ipc_message * msg_p)
{
   ipc_queue_id_t return_qID;
   char * msgData;
   char * str_p;
   char fileName[SMB_DLM_PATH_LENGTH + 3];
   /* out_string holds remote path + local path +file name + internal temp path; #50 is for adding more cd for path segmentations*/
   char put_string[SMB_DLM_PATH_LENGTH + SMB_DLM_FILE_LENGTH * 2 + 50];	
   int  errorCode;

   /*
    * Get the local and remote paths
    */
   LOG_SMB_ENTRY;
   return_qID = ipc_get_message_source( msg_p );
   msgData = (char *)ipc_message_data(msg_p);
   strcpy( session.localPath, msgData);
   str_p = msgData + strlen(msgData) + 1;  /*Get the next string*/
   strcpy(session.remotePath, str_p);
   LOG_SMB_DEBUG("%s(): Local Path = \"%s\"", __FUNC__, session.localPath);
   LOG_SMB_DEBUG("%s(): Remote Path = \"%s\"",__FUNC__, session.remotePath);
   pathCheck(session.localPath);  /* eliminate any quotation marks.*/
   pathCheck(session.remotePath);
   backslash_to_slash(session.localPath); /* In case the seperator is wrong*/
   slash_to_backslash(session.remotePath); /* In case the seperator is wrong*/

   LOGINFO("%s(): Local Path = \"%s\"", __FUNC__, session.localPath);
   LOGINFO("%s(): Remote Path = \"%s\"",__FUNC__, session.remotePath);


    /*
     * construct the put string
     */

   if( isEmpty(session.localPath))
   {
      /* The local path is empty. This is an error!*/
      LOG_SMB_ERROR("%s(): The local path is blank!", __FUNC__);
      errorCode = SMB_DLM_PARAMETER_ERROR;
      ipcData = (ALIGN_TYPE)errorCode;
      ipc_send_message( return_qID, get_send_handle(), &ipcData,
                        sizeof(ipcData), SMB_DLM_PUT_FILE_REPLY);
      LOG_SMB_RETURN;
   }

   if(isEmpty( session.remotePath ))
   {
      if( (str_p = strrchr(session.localPath, '/')) == NULL)
      {
         /* There is no path in local file name */
         strcpy(session.remotePath, session.localPath);
      }
      else
      {
         /* There is a path, copy filename into remote path */
         strcpy(session.remotePath, (str_p + 1));
      }

      checkForSpaces(session.localPath);
      checkForSpaces(session.remotePath);
      strcpy(put_string, "put ");
      strcat(put_string, session.localPath);
      strcat(put_string, " ");
      strcat(put_string, session.remotePath);

   }

   /* Is there a directory listing in the remote path? */
   else  if((str_p = strrchr(session.remotePath, '\\')) == NULL)
   {
      /* There is no directory listing */
      checkForSpaces(session.localPath);
      checkForSpaces(session.remotePath);
      strcpy(put_string, "put ");
      strcat(put_string, session.localPath);
      strcat(put_string, " ");
      strcat(put_string, session.remotePath);
   }
   else
   {
      /*
       * There is a directory listing. Seperate out the file name and check on
       * the size of the path. If greater than 127 characters, break up the
       * the path into segments less than 127 characters.
       */
      str_p = str_p + 1;
      strcpy(fileName, str_p); /* seperate out the filename */
      *str_p = '\0';
      checkForSpaces(session.localPath);
      checkForSpaces(fileName);
      /*
        * check to see if cd command is needed and return with any cd command
        * incorporated into the start of the put_string.
        */
      checkForSpaces(session.remotePath);
      
      if(!cdCommand( put_string, session.remotePath))
      {
         LOG_SMB_ERROR("Error on creating cd command");
         ipcData = (ALIGN_TYPE)SMB_DLM_PARAMETER_ERROR;
         ipc_send_message( return_qID, get_send_handle(), &ipcData,
                          sizeof(ipcData), SMB_DLM_PUT_FILE_REPLY);
         LOG_SMB_RETURN;
      }
      strcat(put_string, "put ");
      strcat(put_string, session.localPath);
      strcat(put_string, " ");
      strcat(put_string, fileName);
      
      LOGFORCE("%s(): put_string = \"%s\"", __FUNC__, put_string);
	  LOGFORCE("%s(): local path = %s, Remote Path = \"%s\"",__FUNC__, session.localPath, session.remotePath);


   }


   /*
    * call smbclient
    */

   /*
    * For Debugging, provide  redirect file
    */
#if DEBUG
   strcpy(session.outFilePath, "/var/log/nc/putDebug");
#else
   session.outFilePath[0] = '\0';
#endif

   errorCode = callSMBclient( put_string);

   ipcData = (ALIGN_TYPE)errorCode;
   ipc_send_message( return_qID, get_send_handle(), &ipcData, sizeof(ipcData),
                     SMB_DLM_PUT_FILE_REPLY);



   /* Clean up paths */
   session.remotePath[0] = '\0';
   session.localPath[0] = '\0';
#if DEBUG
   session.outFilePath[0] = '\0';
#endif
   LOG_SMB_RETURN;
}


/*****************************************************************************

* FUNCTION: smb_filing_getFile
*
* DESCRIPTION: This function copies a file from the SMB file server to the client
*
*
*
*
* PARAMETERS:  Ipc_message * msg_p  -- a pointer to the IPC message body
*
*
* RETURNS:     void
*****************************************************************************/
void  smb_filing_getFile( Ipc_message * msg_p)
{
   ipc_queue_id_t return_qID;
   char * msgData;
   char * str_p;
   char get_string[SMB_DLM_PATH_LENGTH + 2*SMB_DLM_FILE_LENGTH + 10 ];	// was 549,SMB_DLM_PATH_LENGTH(remote) + 2*SMB_DLM_FILE_LENGTH localPath & filename+ 10  
   int  errorCode;

   /*
    * Get the local and remote paths
    */
   LOG_SMB_ENTRY;
   return_qID = ipc_get_message_source( msg_p );
   msgData = (char *)ipc_message_data(msg_p);
   strcpy( session.localPath, msgData);
   str_p = msgData + strlen(msgData) + 1;  /*Get the next string*/
   strcpy(session.remotePath, str_p);
   LOG_SMB_DEBUG("%s(): Remote Path = \"%s\"",__FUNC__, session.remotePath);
   LOG_SMB_DEBUG("%s(): Local path = \"%s\"",__FUNC__, session.localPath);
   pathCheck(session.localPath);  /* remove any quotation marks */
   pathCheck(session.remotePath);
   backslash_to_slash(session.localPath); /* In case the seperator is wrong*/
   slash_to_backslash(session.remotePath); /* In case the seperator is wrong*/


    /*
    * construct the get string
    */


   if(isEmpty( session.remotePath))
   {
      /* Remote path is blank. This is an error! */
      LOG_SMB_ERROR("%s(): The remote path is blank!", __FUNC__);
      errorCode = SMB_DLM_PARAMETER_ERROR;
      ipcData = (ALIGN_TYPE)errorCode;
      ipc_send_message( return_qID, get_send_handle(), &ipcData,
                        sizeof(ipcData), SMB_DLM_GET_FILE_REPLY);
      LOG_SMB_RETURN;
   }

   if(isEmpty( session.localPath ))
   {
      if( (str_p = strrchr(session.remotePath, '\\')) != NULL)
      {
         /* There is a path in the remote file name. It is necessary to copy
         *  the filename into the local path to avoid an error
         */
         strcpy(session.localPath, (str_p + 1));

      }
      checkForSpaces(session.remotePath);
      checkForSpaces(session.localPath);
      strcpy(get_string, "get ");
      strcat(get_string, session.remotePath);
      strcat(get_string, " ");
      if(!isEmpty( session.localPath)) /* copy only if something is there */
         strcat(get_string, session.localPath);
   }
   else
   {
      checkForSpaces(session.remotePath);
      checkForSpaces(session.localPath);

      sprintf(get_string, "get %s %s", session.remotePath, session.localPath);
   }

   /*
   * Construct the command string
   */

   /*
    * For Debugging, provide  redirect file
    */
#if DEBUG
   strcpy(session.outFilePath, "/var/log/nc/getDebug");
#else
   session.outFilePath[0] = '\0';
#endif

   errorCode = callSMBclient( get_string );

   ipcData = (ALIGN_TYPE)errorCode;
   ipc_send_message( return_qID, get_send_handle(), &ipcData, sizeof(ipcData),
                     SMB_DLM_GET_FILE_REPLY);


   /* Clean up paths */
   session.remotePath[0] = '\0';
   session.localPath[0] = '\0';
#if DEBUG
   session.outFilePath[0] = '\0';
#endif
   LOG_SMB_RETURN;
}

/*****************************************************************************

* FUNCTION: smb_filing_listDir
*
* DESCRIPTION: The function gets a directory listing from the SMB file server
*              and places the listing in a file.
*
*
*
* PARAMETERS: Ipc_message * msg_p -- a pointer to the ipc message containing
*                                    the directory to obtain the listing from
*                                    listing and the name of the file to
*                                    redirect output to.
*
*
* RETURNS:    void
*
*****************************************************************************/
void smb_filing_listDir( Ipc_message * msg_p)
{
   ipc_queue_id_t return_qID;
   char * msgData;
   char * str_p;
   /* list_string may hold remote path + local path +file name + internal temp path; #50 is for adding more cd for path segmentations and temp path*/
   char list_string[SMB_DLM_PATH_LENGTH + 2*SMB_DLM_FILE_LENGTH + 50];	//was 549];
   int  errorCode;

   LOG_SMB_ENTRY;

   /*
    * Get the local and remote paths
    */

   return_qID = ipc_get_message_source( msg_p );
   msgData = (char *)ipc_message_data(msg_p);
   strcpy( session.remotePath, msgData); /* Get the remote directory */
   str_p = msgData + strlen(msgData) + 1;  /*Get the next string*/
   strcpy(session.outFilePath, str_p);
   LOG_SMB_DEBUG("%s():Directory to list = \"%s\"", __FUNC__,
                 session.remotePath);
   LOG_SMB_DEBUG("%s(): Redirect File = \"%s\"", __FUNC__,
                  session.outFilePath);
   LOGFORCE("%s():Directory to list = \"%s\", number of UTF8 char %d", __FUNC__,
                 session.remotePath, UTF8_strlen(session.remotePath));
   LOGFORCE("%s(): Redirect File = \"%s\", number of char = %d", __FUNC__,
                  session.outFilePath, UTF8_strlen(session.outFilePath));

   pathCheck(session.outFilePath);  /* remove any quotation marks */
   pathCheck(session.remotePath);
   backslash_to_slash(session.outFilePath); /* In case the seperator is wrong*/
   slash_to_backslash(session.remotePath); /* In case the seperator is wrong*/

    /*
    * construct the list string. We must cd to the remote directory and
    * then issue the ls command.
    */
   if(isEmpty( session.remotePath ))
   {
      /* if this is empty, then get the root directory */
      strcpy(list_string, "ls");
   }
   else
   {
      checkForSpaces(session.remotePath);
      if(!cdCommand(list_string, session.remotePath))
      {
         LOG_SMB_ERROR("Error on creating cd command");
         ipcData = (ALIGN_TYPE)SMB_DLM_PARAMETER_ERROR;
         ipc_send_message( return_qID, get_send_handle(), &ipcData,
                          sizeof(ipcData), SMB_DLM_LIST_DIRECTORY_REPLY);
         LOG_SMB_RETURN; /* return; */
      }
      strcat(list_string, "ls");
   }


   /*
   * Construct the command string
   */

   errorCode = callSMBclient( list_string );

   ipcData = (ALIGN_TYPE)errorCode;
   ipc_send_message( return_qID, get_send_handle(), &ipcData, sizeof(ipcData),
                     SMB_DLM_LIST_DIRECTORY_REPLY);


   /* Clean up paths */
   session.remotePath[0] = '\0';
   session.outFilePath[0] = '\0';
   LOG_SMB_RETURN;
}

/*****************************************************************************

* FUNCTION: smb_filing_deleteFile
*
* DESCRIPTION: This function removes a file from the remote file server.
*
*
*
*
* PARAMETERS:  Ipc_message * msg_p -- a pointer to the IPC message
*
*
* RETURNS:     void
*****************************************************************************/
void  smb_filing_deleteFile( Ipc_message * msg_p)
{
   ipc_queue_id_t return_qID;
   char * msgData;
   char * str_p;
   char fileName[SMB_DLM_PATH_LENGTH + 3];
   char delete_string[300 + SMB_DLM_PATH_LENGTH];  // was 300
   int  errorCode;

    LOG_SMB_ENTRY;

   /*
    * Get the remote paths
    */

   return_qID = ipc_get_message_source( msg_p );
   msgData = (char *)ipc_message_data(msg_p);
   strcpy( session.remotePath, msgData);
   LOG_SMB_DEBUG("%s(): File to delete = \"%s\"", __FUNC__,
                 session.remotePath);
   pathCheck(session.remotePath); /* remove any quotation marks */
   slash_to_backslash(session.remotePath); /* In case the seperator is wrong*/


   /*
   * construct the delete string checking for directories in the path
   */
   delete_string[0] = '\0';
   if((str_p = strrchr(session.remotePath, '\\'))!= NULL)
   {
      /*
      * There is a directory listing. Separate out the file name and check on
      * the size of the path. If greater than 127 characters, break up the
      * the path into segments less than 127 characters.
   	  */
       str_p = str_p + 1;
       strcpy(fileName, str_p); /* seperate out the filename */
       *str_p = '\0';
       checkForSpaces(fileName);
      /*
      * check to see if cd command is needed and return with any cd command
      * incorporated into the start of the delete_string.
      */
      checkForSpaces(session.remotePath);
      if(!cdCommand( delete_string, session.remotePath))
      {
         LOG_SMB_ERROR("Error on creating cd command");
         ipcData = (ALIGN_TYPE)SMB_DLM_PARAMETER_ERROR;
         ipc_send_message( return_qID, get_send_handle(), &ipcData,
                          sizeof(ipcData), SMB_DLM_DELETE_FILE_REPLY);
         LOG_SMB_RETURN; /* return; */
      }
      strcat(delete_string, "del ");
      strcat(delete_string, fileName);

   }
   else
   {
      checkForSpaces(session.remotePath);
      strcpy(delete_string, "del ");
      strcat(delete_string, session.remotePath);
   }

   /*
   * Call smbclient
   */

   /*
    * For Debugging, provide  redirect file
    */
#if DEBUG
   strcpy(session.outFilePath, "/var/log/nc/deleteDebug");
#else
   session.outFilePath[0] = '\0';
#endif

   errorCode = callSMBclient( delete_string );

   ipcData = (ALIGN_TYPE)errorCode;
   ipc_send_message( return_qID, get_send_handle(), &ipcData, sizeof(ipcData),
                     SMB_DLM_DELETE_FILE_REPLY);


   /* Clean up paths */
   session.remotePath[0] = '\0';
#if DEBUG
   session.outFilePath[0] = '\0';
#endif
   LOG_SMB_RETURN ;
}



/*****************************************************************************

* FUNCTION: smb_filing_removeDir
*
* DESCRIPTION: This function removes a directory on the SMB file Server.
*
* PARAMETERS:  Ipc_message * msg_p -- a pointer to the ipc message that
*                                     contains the directory to remove
*                                     including the full path
*
* RETURNS:     void
*
*****************************************************************************/
void smb_filing_removeDir(Ipc_message * msg_p)
{
   ipc_queue_id_t return_qID;
   char * msgData;
   char directoryName[SMB_DLM_PATH_LENGTH + 3];
   char remove_string[300 + SMB_DLM_PATH_LENGTH];  //was 300
   int  errorCode;

   LOG_SMB_ENTRY;

   /*
    * Get the remote directory path
    */

   return_qID = ipc_get_message_source( msg_p );
   msgData = (char *)ipc_message_data(msg_p);
   strcpy( session.remotePath, msgData);
   LOG_SMB_DEBUG("%s(): Directory to delete = \"%s\"", __FUNC__,
                 session.remotePath);
   pathCheck(session.remotePath);  /* remove any quotation marks */
   slash_to_backslash(session.remotePath); /* In case the seperator is wrong*/

   /*
    * Check to see if there are multiple directories in the path.
    */
   if(noPath( directoryName, session.remotePath))
   {
      /* There is no Path */
      checkForSpaces(session.remotePath);
      sprintf(remove_string, "rmdir %s", session.remotePath);
   }
   else
   {
      /* There is a path,  */
      checkForSpaces(directoryName);
      checkForSpaces(session.remotePath);
      if(!cdCommand(remove_string, session.remotePath))
      {
         LOG_SMB_ERROR("Error on creating cd command");
         ipcData = (ALIGN_TYPE)SMB_DLM_PARAMETER_ERROR;
         ipc_send_message( return_qID, get_send_handle(), &ipcData,
                          sizeof(ipcData), SMB_DLM_REMOVE_DIRECTORY_REPLY);
         LOG_SMB_RETURN;
      }
      strcat(remove_string, "rmdir ");
      strcat(remove_string, directoryName);

   }
   /*
    * For Debugging, provide  redirect file
    */
#if DEBUG
   strcpy(session.outFilePath, "/var/log/nc/debugRmDir");
#else
   session.outFilePath[0] = '\0';
#endif
   errorCode = callSMBclient(remove_string);
   ipcData = (ALIGN_TYPE)errorCode;
   ipc_send_message( return_qID, get_send_handle(), &ipcData, sizeof(ipcData),
                     SMB_DLM_REMOVE_DIRECTORY_REPLY);


   /* Clean up paths */
   session.remotePath[0] = '\0';
#if DEBUG
   session.outFilePath[0] = '\0';
#endif
   LOG_SMB_RETURN;
}

/*****************************************************************************

* FUNCTION: smb_filing_makeDir
*
* DESCRIPTION: This function removes a directory on the SMB file Server.
*
* PARAMETERS:  Ipc_message * msg_p -- a pointer to the ipc message that
*                                     contains the parente directory and the
*                                     new child directory.
*
* RETURNS:     void
*
*****************************************************************************/
void smb_filing_makeDir( Ipc_message * msg_p)
{
   ipc_queue_id_t return_qID;
   char * msgData;
   char * str_p;
   char make_string[300 + SMB_DLM_PATH_LENGTH];  //was 300
   int  errorCode;

   LOG_SMB_ENTRY;

   /*
    * Get the parent directory and the new child directory
    */
   return_qID = ipc_get_message_source( msg_p );
   msgData = (char *)ipc_message_data(msg_p);
   strcpy( session.localPath, msgData); /* get the new child directory */
   str_p = msgData + strlen(msgData) + 1;
   strcpy(session.remotePath, str_p); /*Get the parent directory*/
   pathCheck(session.localPath);  /* remove any quotation marks */
   pathCheck(session.remotePath);
   LOG_SMB_DEBUG("%s(): New Directory = \"%s\"",__FUNC__, session.localPath);
   LOG_SMB_DEBUG("%s(): Parent Directory = \"%s\"",__FUNC__,
                 session.remotePath);

   checkForSpaces(session.remotePath);
   checkForSpaces(session.localPath);
   slash_to_backslash(session.localPath); /* In case the seperator is wrong*/
   slash_to_backslash(session.remotePath); /* In case the seperator is wrong*/

   if(isEmpty(session.localPath))
   {
      /* This will create a blank directory. This is an error */
      LOG_SMB_ERROR("%s(): Directory to create is blank", __FUNC__);
      ipcData = (ALIGN_TYPE)SMB_DLM_PARAMETER_ERROR;
      ipc_send_message( return_qID, get_send_handle(), &ipcData,
                         sizeof(ipcData), SMB_DLM_MAKE_DIRECTORY_REPLY);
      LOG_SMB_RETURN;
   }

   if(isEmpty(session.remotePath))
   {
      /* assume parent directory is the current directory */
      sprintf(make_string, "mkdir %s", session.localPath);
   }
   else
   {
      if(!cdCommand(make_string, session.remotePath))
      {
         LOG_SMB_ERROR("Error on creating cd command");
         ipcData = (ALIGN_TYPE)SMB_DLM_PARAMETER_ERROR;
         ipc_send_message( return_qID, get_send_handle(), &ipcData,
                          sizeof(ipcData), SMB_DLM_MAKE_DIRECTORY_REPLY);
         LOG_SMB_RETURN;
      }
      strcat(make_string, "mkdir ");
      strcat(make_string, session.localPath);
   }
   /*
    * For Debugging, provide  redirect file
    */
#if DEBUG
   strcpy(session.outFilePath, "/var/log/nc/debugMkDir");
#else
   session.outFilePath[0] = '\0';
#endif
   errorCode = callSMBclient(make_string);
   ipcData = (ALIGN_TYPE)errorCode;
   ipc_send_message(return_qID, get_send_handle(), &ipcData, sizeof(ipcData),
                     SMB_DLM_MAKE_DIRECTORY_REPLY);


   /* Clean up paths */
   session.remotePath[0] = '\0';
#if DEBUG
   session.outFilePath[0] = '\0';
#endif
   LOG_SMB_RETURN;
}

/*****************************************************************************

* FUNCTION: cdCommand
*
* DESCRIPTION: This checks the path length provided and determines if it
*              exceeds 127 characters (including quotes). If less then 128
*              characters, it returns a cd command using the entire path. If
*              the path is greater than 128 characters, it attempts to parse
*              into segements less than 128 characters. If this is possible,
*              it returns TRUE, FALSE if not possible.
*              Note: checkForSpaces() must be called on the string passed to
*              this function. This function expects strings containing
*              spaces to enclosed in quotes.
*
* PARAMETERS:  char * cd_string -- a string in which to place the cd command
*                                  script
*              char * path -- The path to be parsed.
*
* RETURNS:     TRUE if no errors in path segment sizes
*              FALSE if a path segment exceeds 127 characters
*
*****************************************************************************/
bool_t cdCommand( char * cd_string, char * path)
{
   LOG_SMB_ENTRY;
   LOG_SMB_DEBUG("%s(): path being parsed = \"%s\", # characters = %d", __FUNC__, path, UTF8_strlen(path));

   cd_string[0] = '\0';

   if(isEmpty(path))
   {
      SESS_RETURN3(TRUE, "%s() path is empty, nothing to do,"
                   "returning TRUE", __FUNC__);
      /* return TRUE;  nothing do do */
   }

   sprintf( cd_string, "cd %s;", path);
   LOGINFO("%s(): cd_string = %s, # characters = %d", __FUNC__, cd_string, UTF8_strlen(cd_string));

   SESS_RETURN3(TRUE, "%s(): Returning full path",
                __FUNC__);
}






/*****************************************************************************

* FUNCTION: checkForSpaces
*
* DESCRIPTION: This function checks to see if the path provided contains any
*              spaces. If it does, the entire path is enclosed in quotes.
*              Note: pathCheck() should be called first to eliminate any
*              quotation marks in the path string.
*
*

* PARAMETERS:  char * path -- path to check
*
* RETURNS:  void
*
*****************************************************************************/
static void checkForSpaces( char * path)
{
   char * char_p;
   size_t strLength = 0;

   LOG_SMB_ENTRY;
   LOG_SMB_DEBUG("%s(): checking path = \"%s\"", __FUNC__, path);


   if((char_p = strchr(path, ' '))!= NULL && (strLength = strlen(path)) > 0)
   {
      /*append a " */
      LOG_SMB_DEBUG("%s(): This path contains spaces! Adding quotes",
                     __FUNC__);
      path[strLength + 1] = '\0';
      path[strLength] = '"';
      memmove(path+1, path, strLength + 2);
      path[0] = '"';
      LOG_SMB_DEBUG("%s(): path after checking spaces = \"%s\"",
                     __FUNC__, path);
   }
   else
   {
     LOG_SMB_DEBUG("%s(): This path contains no spaces", __FUNC__);
   }

   LOG_SMB_RETURN;
}


/*****************************************************************************

* FUNCTION: callSMBclient
*
* DESCRIPTION: This function creates a command string and calls the SAMBA
*              client via a system call. It checks the length of the command
*              string, if it exceed MAX_COMMAND_LENGTH, then it returns an
*              Error. It returns to the caller the value returned by the
*              system call.
*
*
*
* PARAMETERS:  char * smbCommands -- a string containing the commands to be
*                                    performed by the smbclient after it is
*                                    is called.
*
* RETURNS:  int -- the value returned by the system call or an error code
*                  if the command string is too big.
*
*****************************************************************************/
static int callSMBclient( char * smbCommands)
{
	char *dlm_dir = XPATH_NC_DLMS_DIR;
	char *smb_client_path = "smb/apps/smbclient";
	size_t client = strlen(dlm_dir) + strlen(smb_client_path);
	size_t uString = strlen(session.Ustring);
	size_t config = strlen(SMB_DLM_CONFIG_FILE);	/* /etc/smb.conf.file */
	size_t smbCmd = strlen(smbCommands);
	size_t outFile = strlen(session.outFilePath);
	size_t port = 0;
	size_t commandLength = 0;
	int retVal = 0;
	size_t addChar = 12;
	char outFileStrg[SMB_DLM_PATH_LENGTH + 5]; /* includes -d3> */
	char portStrg[10];
	char commandString[MAX_COMMAND_LENGTH + 1];

	/*
	 * The command line will look like this when constructed:
	 *
	 * "smbclient service Ustring -sconfigFile -c'smbCommands' port outfile"
	 *
	 * Additional charcters not included in the string lengths are: 6 spaces
	 * and the characters -,s,-,c,','.  Therefore addChar = 6+6 =12
	 */

  LOG_SMB_ENTRY;

  /* check to see if the port is set to default. If so, do not need to
   * to include the port number command.
   */
  if( session.port == SMB_DLM_DEFAULT_PORT)
  {
      portStrg[0] = '\0';
  }
  else
  {
     sprintf(portStrg, "-p%d", session.port);
     port = strlen(portStrg); 
  }

	/* check to see if there is an outfile, If so, set debug level to 3
	 * and redirect output to this file
	 */
	if(outFile != 0)
	{
#if DEBUG
		/* Dump debug information */
		sprintf(outFileStrg, "-d5 >> %s 2>&1",session.outFilePath);
#else
		sprintf(outFileStrg, "> %s 2>&1", session.outFilePath);
#endif
	}
	else
	{
        sprintf(outFileStrg, "> /dev/null 2>&1");
	}

	outFile = strlen(outFileStrg);

	/* check the length */
	commandLength = client  + uString + config + smbCmd + port
		+ outFile + addChar;
	
	LOG_SMB_DEBUG("commandLength = %d, client= %d, uString = %d, config = %d, smbCmd = %d, port=%d, outFile = %d, addChar=%d", 
					commandLength, client, uString, config, smbCmd, port, outFile,addChar);

	if(commandLength >  MAX_COMMAND_LENGTH)
	{
		retVal = SMB_DLM_INTERNAL_ERROR;
		LOG_SMB_ERROR("Command Line exceeds %d characters!"
				,MAX_COMMAND_LENGTH);
	}
	else
	{

		switch (session.krbFlngMode)
		{
			case KerberosAuthorizationPreferred:

				setKerberosCacheLocationWithFilename(session.kerberosCachePath);

				LOG_SMB_DEBUG("Kerberos try mode detected.  Now doing kerbetized samba client call");

				sprintf( commandString, "%s%s -k -s%s -c'%s' %s %s",
						dlm_dir, smb_client_path, SMB_DLM_CONFIG_FILE,
						smbCommands, portStrg, outFileStrg);

#if COMMANDDEBUG
				/* Avoiding printing out the password to the smbipc log. Print only if
				 * necessary for debugging purposes.
				 */
				LOGFORCE("commandString = %s\n", commandString);
#endif

				retVal = (int)system(commandString);

				/* For linux, the return value from system() will be in a
				 * different format, the actual return value is offset by 8 bits. Use
				 * sys/wait macros to make the conversion.
				 */
					
				if(retVal == 0)	
				{
					LOG_SMB_INFO("Kerbetized samba succeeded!.  Changing kerberos mode from KerberosAuthorizationPreferred to KerberosAuthorizationAlways");
		   			session.krbFlngMode = KerberosAuthorizationAlways;
				}
				else
				{
					LOG_SMB_INFO("Kerbetized samba failed.  Changing kerberos mode from KerberosAuthorizationPreferred to KerberosAuthorizationOff");
					session.krbFlngMode = KerberosAuthorizationOff;
				}

				
				if ( retVal != 0 )
				{
					LOG_SMB_WARN("Kerbetized samba client call failed.");

					LOG_SMB_DEBUG("Using samba client call with username and password.");

					sprintf( commandString , "%s%s %s -s%s -c'%s' %s %s",
							dlm_dir, smb_client_path, session.Ustring, SMB_DLM_CONFIG_FILE,
							smbCommands, portStrg, outFileStrg);

#if COMMANDDEBUG
					/* Avoiding printing out the password to the smbipc log. Print only if
					 * necessary for debugging purposes.
					 */
					LOGFORCE("commandString = %s\n", commandString);
#endif

				
					retVal = (int)system(commandString);
					
					/* For linux, the return value from system() will be in a
					 * different format, the actual return value is offset by 8 bits. Use
					 * sys/wait macros to make the conversion.
					 */

				}

				break;

			case KerberosAuthorizationAlways:	

				setKerberosCacheLocationWithFilename(session.kerberosCachePath);

				LOG_SMB_DEBUG("Kerberos only mode detected. Now doing kerbetized samba client call");

				sprintf( commandString, "%s%s -k -s%s -c'%s' %s %s",
						dlm_dir, smb_client_path, SMB_DLM_CONFIG_FILE,
						smbCommands, portStrg, outFileStrg);

#if COMMANDDEBUG
				/* Avoiding printing out the password to the smbipc log. Print only if
				 * necessary for debugging purposes.
				 */
				LOGFORCE("commandString = %s\n", commandString);
#endif

				retVal = (int)system(commandString);

				/* For linux, the return value from system() will be in a
				 * different format, the actual return value is offset by 8 bits. Use
				 * sys/wait macros to make the conversion.
				 */

				break;

			case KerberosAuthorizationOff:

				LOG_SMB_DEBUG("Kerberos off mode detected. Using samba client call with username and password.");

				sprintf( commandString, "%s%s %s -s%s -c'%s' %s %s",
						dlm_dir, smb_client_path, session.Ustring, SMB_DLM_CONFIG_FILE,
						smbCommands, portStrg, outFileStrg);


#if COMMANDDEBUG
				/* Avoiding printing out the password to the smbipc log. Print only if
				 * necessary for debugging purposes.
				 */
				LOGFORCE("commandString = %s\n", commandString);
#endif

				retVal = (int)system(commandString);
				
				/* For linux, the return value from system() will be in a
				 * different format, the actual return value is offset by 8 bits. Use
				 * sys/wait macros to make the conversion.
				 */

				break;

			default:
				LOG_SMB_ERROR("Invalid kerberos mode detected.");

				break;


		}

		memset(commandString,0,sizeof(commandString));

#if defined (__linux__)
		if( WIFEXITED(retVal) )
		{
			/* A normal exit */
			retVal = WEXITSTATUS(retVal);
		}
		else
		{
			/* an abnormal exit. Set retVal to internal error */
			LOG_SMB_ERROR("%s(): Abnormal exit from system() call",
					__FUNC__);
			retVal = SMB_DLM_INTERNAL_ERROR;
		}
#endif
	}

	SESS_RETURN4(retVal,"%s(): returns %d", __FUNC__, retVal);
}
/*****************************************************************************

* FUNCTION: isEmpty
*
* DESCRIPTION: This Function checks a path to see if it is empty. Empty is
*              defined as 0 strlength or if it contains only spaces. The
*              function returns TRUE if empty or FALSE otherwise.
*              Note: pathCheck() should be called on the string passed to this
*              function to remove quotation marks before passing the string to
*              this function.
*
*
*
* PARAMETERS:  char * path --  the path to check
*
* RETURNS:  bool_t  -- a boolean
*
*****************************************************************************/
static bool_t isEmpty( char * aString)
{
   size_t strLength = 0;
   bool_t retVal = TRUE;
   unsigned int i, start = 0;

   LOG_SMB_ENTRY;
   LOG_SMB_DEBUG("%s(): string being checked = \"%s\"", __FUNC__, aString);

   if((strLength = strlen(aString)) == 0)
   {
      SESS_RETURN3( TRUE, "%s(): returns TRUE", __FUNC__);
   }


   for( i = start; i < strLength; i++)
   {
      if( aString[i] != ' ' )
      {
         retVal = FALSE;
         break;
      }
   }
   SESS_RETURN4(retVal, "%s(): returns %s", __FUNC__, (retVal)? "TRUE":"FALSE");
}


/*****************************************************************************
* FUNCTION:  clearFilingSessionData
*
* DESCRIPTION: This Function clears the FilingSessionData structure
*              setting the members to their default values or clearing
*              data out of them.
*
*
*
*
* PARAMETERS:  none
*
* RETURNS:  void
*
*****************************************************************************/
static void clearFilingSessionData(void)
{
   session.service[0] = '\0';
   session.Ustring[0] = '\0';
   session.remotePath[0] = '\0';
   session.localPath[0] = '\0';
   session.outFilePath[0] = '\0';
   session.port = SMB_DLM_DEFAULT_PORT;
   session.krbFlngMode = KerberosAuthorizationOff;
}


/*****************************************************************************
* FUNCTION:    noPath
*
* DESCRIPTION: Checks to see if a path contains multiple directories.
*              If multiple directories are found, then there is a path
*              to the last directory listed. If multiple directories are
*              found, it returns the directory leaf in the directoryName
*              string, truncates the remote path and returns a value of FALSE.
*              If only one directory is found the string is blank or null,
*              it returns TRUE  This function should not be called for
*              filename paths.
*
* PARAMETERS:  char * directoryName -- string for the deepest directory
*                                      if found.
*              char * remotePath  -- the remote path to check
*
* RETURNS:     bool_t TRUE if no path is found
*                     FALSE if multiple directories are found.
*
*****************************************************************************/
static bool_t noPath( char * directoryName, char * remotePath)
{
   size_t strLength;
   int   pathPos;
   char * char_p;
   bool_t retVal = TRUE;

   LOG_SMB_ENTRY;

   /* Is this a null string? */
   if((strLength = strlen(remotePath)) == 0)
   {
      /* This is an empty string, nothing to do */
     SESS_RETURN3(TRUE, "%s(): returns TRUE", __FUNC__);
   }
   /* Look for backslashes starting from the end of the path */
   pathPos = strLength - 1;
   while( pathPos > -1 && (remotePath[pathPos] != '\\'))pathPos--;
   if(pathPos == -1)
      retVal = TRUE; /* there is no path indicated */
   else if(pathPos == (int)(strLength - 1))
   {
      /* Look for another slash */
      pathPos--;
      while( pathPos > -1 && (remotePath[pathPos] != '\\'))pathPos--;
      if(pathPos == -1)
         retVal = TRUE;
      else
      {
         retVal = FALSE;
         char_p = &remotePath[pathPos + 1];
         strcpy(directoryName, char_p);
         *char_p = '\0';
      }
   }
   else
   {
      retVal = FALSE;
      char_p = &remotePath[pathPos + 1];
      strcpy(directoryName, char_p);
      *char_p = '\0';
   }
   SESS_RETURN4(retVal,"%s(): returns %s",__FUNC__,(retVal)? "TRUE":"FALSE");
}

/*****************************************************************************
* FUNCTION:    pathCheck
*
* DESCRIPTION: Checks to see if a path contains any quotation marks, leading
*              blank spaces, or trailing blank spaces, If found, they are
*              removed. This is to make processing the strings by other functions
*              easier.
*
* PARAMETERS:  char * path -- the path to check.
* RETURN: void
*****************************************************************************/
static void pathCheck( char * path)
{
   int pathPos = 0, workStrLength = 0, endPos;
   char workingString[SMB_DLM_PATH_LENGTH];
   bool_t begin = TRUE;

   LOG_SMB_ENTRY;

   LOG_SMB_DEBUG("%s(): Checking path = \"%s\"", __FUNC__, path);

   /* Find the end of this string ignoring spaces */
   if((endPos = strlen(path)) == 0)
   {
      /* Empty String,  nothing to do */
      LOG_SMB_RETURN;
   }

   endPos--; /* ignore the null character */

   while( endPos >= 0 && path[endPos] == ' ') endPos--;

   if( endPos == -1 )
   {
      /* This string is only blanks, nothing to do */
      LOG_SMB_RETURN;
   }

   while( pathPos <= endPos)
   {
      if(begin)
      {

         if( path[pathPos] == '"')
         {
            /* Do not copy this character, but start adding characters to
            ** the workingString.
            */
            begin = FALSE;
         }
         else if( path[pathPos] != ' ')
         {
            /* Not a space character or quote, begin adding characters to
            ** workingString.
            */
            begin = FALSE;
            workingString[workStrLength] = path[pathPos];
            workStrLength++;
         }
      }
      else if( path[pathPos] != '"')
      {
         /* If not a quote, add this character to the working string. */
         workingString[workStrLength] = path[pathPos];
         workStrLength++;
      }
      pathPos++;  /* Examine the next character */
   }
   /* terminate the workingString */
   workingString[workStrLength] = '\0';
   strcpy(path, workingString);
   LOG_SMB_DEBUG("%s(): path afterPathCheck = \"%s\"", __FUNC__, path);
   LOG_SMB_RETURN;
}




/*****************************************************************************
* FUNCTION:    slash_to_backslash
*
* DESCRIPTION: Converts all '/' characters to '\'. Called for remote paths.
*
* PARAMETERS:  char * path -- the path to check
*
* RETURNS: void
*****************************************************************************/
static void slash_to_backslash( char * path){
    char * char_p;

    LOG_SMB_ENTRY;
    char_p = path;

    while((char_p = strchr( char_p, '/')) != NULL){
        *char_p = '\\';
    }
    LOG_SMB_DEBUG("%s(): Modified path: \"%s\"", __FUNC__, path);

    LOG_SMB_RETURN;
}


/*****************************************************************************
* FUNCTION:    backslash_to_slash
*
* DESCRIPTION: Converts all '\' characters to '/'. Called for local paths.
*
* PARAMETERS:  char * path -- the path to check
*
* RETURNS: void
*****************************************************************************/
static void backslash_to_slash( char * path){
    char * char_p;

    LOG_SMB_ENTRY;

    char_p = path;
    while((char_p = strchr( char_p, '\\')) != NULL){
        *char_p = '/';
    }

    LOG_SMB_DEBUG("%s(): Modified path: \"%s\"", __FUNC__, path);

    LOG_SMB_RETURN;
}

/*****************************************************************************
* FUNCTION:    doWeUseTheLDAPDomain
*
* DESCRIPTION: This function will look at several network setting and determine
*   if we should be using a domain name found in LDAP settings or not.
*
*   NOTE: THIS IS A BAD WAY TO DO THIS!!!  In reality, we should be making these
*   determinations in the scan service (DistributionService or TransferService for
*   example).  But somehow we ended up doing it in the lower level protocol base
*   which is causing lots of issues.  This function is a bandaid to try and partially
*   fix the real problem that is being seen in the field (in this case, this problem
*   was introduced by CQGbl00357794).
*
* PARAMETERS:  none
*
* RETURNS: TRUE if we should be using a domain name from the LDAP server settings,
*     FALSE if otherwise.
*****************************************************************************/
static bool_t doWeUseTheLDAPDomain()
{
    bool_t useLDAPDomain = FALSE;
    bool_t isLdapAuthEnabled = FALSE;
    bool_t appendDnEnabled = FALSE;
    status_t cfgStatus = X_SUCCESS;

    LOGENTER(__FUNCTION__);

    cfgStatus = Cfg_GetBoolAttrValue(configServiceLdapAuthEnabled, &isLdapAuthEnabled);

    /* sanity check that configServer worked, if not, success if false */
    if (cfgStatus != X_SUCCESS)
    {
        LOGERROR("Value for %s could not be retrieved from the Config Server %d",
                configServiceLdapAuthEnabled , (int) cfgStatus);
        useLDAPDomain = FALSE;
    }
    else
    {

        /*
         * If authentication for LDAP is enabled, then we maybe using the domain configured by
         * the user for LDAP, so check for it.
         ******/
        if (isLdapAuthEnabled == TRUE)
        {

            cfgStatus = Cfg_GetBoolAttrValue(configServerLdapAppendDnArg, &appendDnEnabled);

            /* sanity check configServer worked, if not, success if false */
            if (cfgStatus != X_SUCCESS)
            {
                LOGERROR("Value for %s could not be retrieved from the Config Server %d",
                                         configServerLdapAppendDnArg , (int) cfgStatus);
                useLDAPDomain = FALSE;
            }
            else
            {
                if (appendDnEnabled == TRUE)
                {
                    useLDAPDomain = TRUE;
                }
                else
                {
                    useLDAPDomain = FALSE;
                }
            }
        }
    }
    LOGRETURN(__FUNCTION__);
    return (useLDAPDomain);
}
